from typing import Any
import logging
from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.models import Model


class DefaultService:

    @staticmethod
    @transaction.atomic
    def save_instance(instance: Model, **field_data: Any) -> tuple[Model, bool]:
        """
            Apply field_data to `instance`, validate, and persist.
            Accepts both `field` and `field_id` for FKs.
            Works for unsaved (create) and existing (update) instances.
            Skips editable = False and auto created fields.
        """
        if not field_data:
            return instance, False

        concrete = {
            f.name: f
            for f in instance._meta.get_fields()
            if f.concrete and not f.many_to_many and not f.one_to_many
        }

        fk_id_aliases = {f"{n}_id" for n, f in concrete.items() if f.many_to_one}
        allowed = set(concrete) | fk_id_aliases

        touched: list[str] = []

        for field, value in field_data.items():
            if field not in allowed:
                logging.warning(f'Field {field!r} is not valid for {instance.__class__.__name__}')
                continue

            model_field = concrete.get(field.rstrip("_id"), None)

            # Skip read-only / auto columns
            if model_field and (getattr(model_field, 'auto_created', False) or not model_field.editable):
                continue

            setattr(instance, field, value)
            touched.append(field.rstrip('_id'))

        # Validate only fields we touched
        try:
            instance.full_clean(exclude=[f for f in concrete if f not in touched])
        except ValidationError as exc:
            raise  # bubble up or wrap as needed

        if instance.pk is None:
            instance.save()
            created = True
        else:
            instance.save(update_fields=touched)
            created = False

        return instance, created



class ServiceDescriptor:
    """
        Acts like a django objects manager.
        If an attribute you access on another object is actually a descriptor.
    """
    def __init__(self, service_cls):
        self._service_cls = service_cls

    def __get__(self, instance, owner):
        # When accessed through the class, instance == None
        target = instance if instance is not None else owner
        return self._service_cls(target)



from django.db import models

from django_spire.notification.models import Notification
from django_spire.notification.sms.querysets import SmsNotificationQuerySet


class SmsNotification(models.Model):
    notification = models.OneToOneField(
        Notification,
        editable=False,
        on_delete=models.CASCADE,
        related_name='sms',
        related_query_name='sms',
    )
    to_phone_number = models.CharField(max_length=11, blank=True)

    objects = SmsNotificationQuerySet.as_manager()

    def __str__(self):
        return f'{self.to_phone_number} - {self.notification.title}'

    class Meta:
        db_table = 'django_spire_notification_sms'
        verbose_name = 'SMS Notification'
        verbose_name_plural = 'SMS Notifications'



from django.contrib.auth.models import User
from django.test import TestCase

from django_spire.auth.user.tests.factories import create_super_user
from django_spire.contrib.services.base_service import BaseService
from django_spire.contrib.services.tests.factories import UserService



class TestBaseService(TestCase):
    def setUp(self):
        self.user = create_super_user()

    def test_abstraction_on_init(self):

        class BrokenService(BaseService):
            def create_taco(self):
                return "Taco!"

        with self.assertRaises(NotImplementedError) as context:
            BrokenService()

    def test_is_class_instance(self):
        service = UserService(self.user)
        self.assertTrue(service.is_class_instance(User))
        self.assertFalse(service.is_class_instance(self.user))

    def test_is_ready_instance(self):
        service = UserService(self.user)

        self.assertTrue(service.is_ready_instance(self.user))
        self.assertFalse(service.is_ready_instance(User))

    def test_is_new_instance(self):
        service = UserService(self.user)
        new_user = User()

        self.assertTrue(service.is_new_instance(new_user))
        self.assertFalse(service.is_new_instance(self.user))



from django.contrib.auth.models import User
from django.utils.timezone import now

from django_spire.history.querysets import HistoryQuerySet
from django_spire.notification.choices import (
    NotificationStatusChoices,
    NotificationTypeChoices,
    NotificationPriorityChoices,
)


class NotificationQuerySet(HistoryQuerySet):
    def app_notifications(self):
        return self.filter(type=NotificationTypeChoices.APP)

    def by_user(self, user: User):
        return self.filter(user=user)

    def email_notifications(self):
        return self.filter(type=NotificationTypeChoices.EMAIL)

    def errored(self):
        return self.filter(status=NotificationStatusChoices.ERRORED)

    def failed(self):
        return self.filter(status=NotificationStatusChoices.FAILED)

    def high_priority(self):
        return self.filter(priority=NotificationPriorityChoices.HIGH)

    def low_priority(self):
        return self.filter(priority=NotificationPriorityChoices.LOW)

    def medium_priority(self):
        return self.filter(priority=NotificationPriorityChoices.MEDIUM)

    def pending(self):
        return self.filter(status=NotificationStatusChoices.PENDING)

    def processing(self):
        return self.filter(status=NotificationStatusChoices.PROCESSING)

    def push_notifications(self):
        return self.filter(status=NotificationTypeChoices.PUSH)

    def ready_to_send(self):
        return self.pending().filter(publish_datetime__lte=now())

    def sent(self):
        return self.filter(status=NotificationStatusChoices.SENT)

    def sms_notifications(self):
        return self.filter(type=NotificationTypeChoices.SMS)



from __future__ import annotations

from abc import abstractmethod
from typing import Type

from django.db.models import Model

from django_spire.contrib.services.default_service import DefaultService


class BaseService:

    @abstractmethod
    def __init__(self, *arg):
        raise NotImplementedError('You must pass the model class or model instance to the service constructor.')

    @staticmethod
    def is_ready_instance(obj: Model | Type[Model]):
        from django.db.models import Model
        return isinstance(obj, Model) and obj.id is not None

    @staticmethod
    def is_new_instance(obj : Model | Type[Model]):
        return obj.id is None

    @staticmethod
    def is_class_instance(obj: Model | Type[Model]):
        return isinstance(obj, type) and issubclass(obj, Model)

    @property
    def default(self):
        return DefaultService()



from django.contrib.auth.models import User

from django_spire.contrib.services.base_service import BaseService


class UserService(BaseService):

    def __init__(self, user: User):
        self.user = user




from django.contrib.auth.models import Group, User
from django.core.exceptions import FieldDoesNotExist
from django.test import TestCase

from django_spire.auth.mfa.models import MfaCode
from django_spire.auth.user.tests.factories import create_super_user
from django_spire.contrib.services.default_service import DefaultService


class UpdateServiceTestCase(TestCase):
    def setUp(self):
        self.user = create_super_user()
        self.group = Group.objects.create(name='Bobberts Minions')
        self.default_service = DefaultService()

    def test_valid_update_model_fields(self):
        data = {
            'first_name': 'John',
            'last_name': 'Smith'
        }

        user, created = self.default_service.save_instance(self.user, **data)
        self.assertEqual(user.first_name, 'John')
        self.assertEqual(user.last_name, 'Smith')

    def test_valid_create_model_fields(self):
        data = {
            'first_name': 'John',
            'last_name': 'Smith'
        }

        user, created = self.default_service.save_instance(User(), **data)
        self.assertEqual(user.first_name, 'John')
        self.assertEqual(user.last_name, 'Smith')
        self.assertIsNotNone(user.id)

    def test_invalid_field_name(self):
        # Skips the field and saves the instance.
        data = {
            'invalid_field': 'test'
        }
        self.default_service.save_instance(self.user, **data)

    def test_fk_id_aliases(self):
        mfa = MfaCode.objects.create(
            user=self.user,
            code='123456',
            expiration_datetime='2023-01-01'
        )
        new_user = create_super_user()

        data = {
            'user_id': new_user.id,
        }

        mfa, created = self.default_service.save_instance(mfa, **data)
        self.assertEqual(mfa.user_id, new_user.id)

    def test_obj_fk_id_aliases(self):
        mfa = MfaCode.objects.create(
            user=self.user,
            code='123456',
            expiration_datetime='2023-01-01'
        )
        new_user = create_super_user()

        data = {
            'user': new_user,
        }

        mfa, created = self.default_service.save_instance(mfa, **data)
        self.assertEqual(mfa.user_id, new_user.id)



from __future__ import annotations

from django_spire.notification.managers import NotificationManager


def process_notifications() -> str:
    NotificationManager().process_ready_notifications()

    return 'Successfully Completed'



from django.db import models

from django_spire.notification.models import Notification
from django_spire.notification.email.querysets import EmailNotificationQuerySet


class EmailNotification(models.Model):
    notification = models.OneToOneField(
        Notification,
        editable=False,
        on_delete=models.CASCADE,
        related_name='email',
        related_query_name='email',
    )
    to_email_address = models.EmailField()

    objects = EmailNotificationQuerySet.as_manager()

    def __str__(self):
        return f'{self.notification.title} - {self.to_email_address}'

    class Meta:
        db_table = 'django_spire_notification_email'
        verbose_name = 'Email Notification'
        verbose_name_plural = 'Email Notifications'



from django.contrib import admin

from django_spire.notification.sms.models import SmsNotification


@admin.register(SmsNotification)
class EmailNotificationAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'notification', 'to_phone_number'
    )



import logging

import django_glue as dg

from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect
from django.urls import reverse
from django.views.decorators.http import require_http_methods
from django.apps import apps

from django_spire.core.shortcuts import get_object_or_null_obj
from django_spire.auth.group.decorators import permission_required
from django_spire.contrib.form.utils import show_form_errors
from django_spire.contrib.generic_views import portal_views
from django_spire.help_desk import forms
from django_spire.help_desk.models import HelpDeskTicket
from django_spire.help_desk.notifications import HelpDeskTicketNotifications
from django_spire.notification.app.models import AppNotification
from django_spire.notification.email.models import EmailNotification


@permission_required('django_spire_help_desk.change_helpdeskticket')
def ticket_form_view(request, pk: int = 0):
    ticket = get_object_or_null_obj(HelpDeskTicket, pk=pk)

    dg.glue_model_object(request, 'ticket', ticket)

    if request.method == 'POST':
        form = forms.HelpDeskTicketForm(request.POST, instance=ticket)

        if form.is_valid():
            ticket = form.save(user=request.user)
            HelpDeskTicketNotifications(ticket).send_ticket_created_notifications()

            return redirect(reverse('django_spire:help_desk:page:list'))

        show_form_errors(request, form)

    else:
        form = forms.HelpDeskTicketForm(instance=ticket)

    return portal_views.model_form_view(
        request=request,
        obj=ticket,
        template='django_spire/help_desk/page/ticket_form_page.html',
        form=form,
        context_data={'ticket': ticket}
    )